// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

extern crate dgl;
extern crate dwindow;
extern crate dgl_gl as gl;

use std::thread;
use std::time::{Duration, Instant};

use dgl::DisplayExt;
use gl::types::*;

const VERSION: dgl::Version = dgl::Version(1, 1);

fn main () {
    // Open the dwindow display.
    let display = dwindow::Display::open().expect("can't open display");

    // Choose an OpenGL surface configuration.
    let config = display.choose_gl_config(&dgl::ConfigInfo {
        supports_gl: true,
        draw_to_window: true,
        ..Default::default()
    }).expect("can't choose OpenGL config");

    // Create a window to render on.
    let window = display.new_window()
                        .with_visual(&config)
                        .build().expect("can't create window");

    // Create an OpenGL context.
    println!("Requesting OpenGL version {}...", VERSION);

    let context = display.new_gl_context()
                         .with_config(&config)
                         .with_version(VERSION)
                         .build().expect("can't create OpenGL context");

    // Initialize the OpenGL state.
    let gl = context.with(&window, |cur| {
        // Check the acquired OpenGL context version.
        match cur.version() {
            Ok(ver) => { println!("Using OpenGL version {}.", ver); },
            Err(err) => { println!("Can't determine OpenGL version: {}", err); },
        }

        // Try to enable vsync.
        match cur.set_swap_interval(-1) {
            Ok(()) => { println!("Vsync enabled."); },
            Err(err) => {
                println!("Can't enable vsync: {}", err);
                println!("Using frame limiting instead.");
            },
        };

        // Load OpenGL functions.
        let gl = gl::Gl::load_with(|name| {
            cur.get_proc_address(name).expect("can't load OpenGL functions")
        });

        // Additional setup.
        unsafe {
            gl.ClearColor(0.0, 0.0, 0.0, 0.0);
        }

        gl
    }).expect("can't select OpenGL context");

    // Show the window.
    window.set_visible(true).expect("can't show window");

    // Set up frame limiting.
    let interval = Duration::from_millis(10); // Limit to 100 fps
    let mut time = Instant::now();
    let mut sleep_until = time + interval;

    // Run the main loop.
    display.run_sync(|event| {
        match event {
            // Stop the application.
            dwindow::Event::Close(_) | dwindow::Event::Destroy(_) => { display.stop(); },

            // Render the next frame.
            dwindow::Event::Update(_) => {
                context.with(&window, |cur| {
                    let size = window.size();

                    unsafe {
                        gl.Viewport(0, 0, size.x as GLsizei, size.y as GLsizei);
                        gl.Clear(gl::COLOR_BUFFER_BIT);

                        gl.Begin(gl::TRIANGLES);
                        gl.Color3f(1.0, 0.0, 0.0);
                        gl.Vertex2f(-0.8, -0.8);
                        gl.Color3f(0.0, 1.0, 0.0);
                        gl.Vertex2f(0.8, -0.8);
                        gl.Color3f(0.0, 0.0, 1.0);
                        gl.Vertex2f(-0.8, 0.8);
                        gl.End();
                    }

                    cur.swap_buffers().expect("can't swap buffers");
                }).expect("can't select OpenGL context");

                // Limit the frame rate.
                time = Instant::now();

                if time < sleep_until {
                    thread::sleep(sleep_until - time);
                } else {
                    while time > sleep_until {
                        sleep_until += interval;
                    }
                }
            },

            _ => ()
        }
    }).expect("can't poll events");
}
