// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::cell::RefCell;
use std::ffi::CStr;
use std::fmt::{self, Formatter};
use std::mem;
use std::os::raw::*;
use std::str::{self, FromStr};

use dwindow::Display;

use config::Config;
use error::Result;
use provider::{ContextProvider, CurrentProvider, GetProvider};
use surface::SurfaceRef;

thread_local!(static THREAD_LOCKED: RefCell<bool> = RefCell::new(false));

/// OpenGL API.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Api {
    Opengl,
    Opengles,
}

/// OpenGL profile.
#[derive(Clone, Copy, Debug, Eq, PartialEq)]
pub enum Profile {
    Core,
    Compatibility,
}

/// OpenGL version.
#[derive(Clone, Copy, Eq, Ord, PartialEq, PartialOrd)]
pub struct Version (pub u8, pub u8);

impl From<(u8, u8)> for Version {
    fn from (v: (u8, u8)) -> Version { Version(v.0, v.1) }
}

impl fmt::Debug for Version {
    fn fmt (&self, f: &mut Formatter) -> fmt::Result {
        fmt::Display::fmt(self, f)
    }
}

impl fmt::Display for Version {
    fn fmt (&self, f: &mut Formatter) -> fmt::Result {
        write!(f, "{}.{}", self.0, self.1)
    }
}

/// OpenGL context builder.
pub struct ContextBuilder<'a> {
    api: Api,
    config: Option<Config>,
    display: &'a Display,
    profile: Option<Profile>,
    share: Option<&'a Context>,
    version: Option<Version>,
}

impl<'a> ContextBuilder<'a> {
    pub fn api (&self) -> Api { self.api }

    pub fn build (&self) -> Result<Context> {
        let provider = ContextProvider::new(self)?;
        Ok(Context { provider })
    }

    pub fn config (&self) -> Option<&Config> { self.config.as_ref() }
    pub fn display (&self) -> &'a Display { self.display }

    #[doc(hidden)]
    pub fn new (display: &'a Display, api: Api) -> ContextBuilder<'a> {
        ContextBuilder {
            api: api,
            config: None,
            display: display,
            profile: None,
            share: None,
            version: None,
        }
    }

    pub fn profile (&self) -> Option<Profile> { self.profile }
    pub fn shared_context (&self) -> Option<&'a Context> { self.share }
    pub fn version (&self) -> Option<Version> { self.version }

    pub fn with_compatibility_profile (mut self) -> ContextBuilder<'a> {
        self.profile = Some(Profile::Compatibility);
        self
    }

    pub fn with_config<T: AsRef<Config>> (mut self, config: T) -> ContextBuilder<'a> {
        self.config = Some(config.as_ref().clone());
        self
    }

    pub fn with_core_profile (mut self) -> ContextBuilder<'a> {
        self.profile = Some(Profile::Core);
        self
    }

    pub fn with_profile (mut self, profile: Profile) -> ContextBuilder<'a> {
        self.profile = Some(profile);
        self
    }

    pub fn with_shared_lists<'b, 'c> (self, share: &'b Context) -> ContextBuilder<'c>
        where 'a: 'c, 'b: 'c
    {
        ContextBuilder {
            api: self.api,
            config: self.config,
            display: self.display,
            profile: self.profile,
            share: Some(share),
            version: self.version,
        }
    }

    pub fn with_version<V: Into<Version>> (mut self, version: V) -> ContextBuilder<'a> {
        self.version = Some(version.into());
        self
    }
}

/// `Context` provider trait.
pub trait IContextProvider<'a> : Sized + Eq {
    type Current : ICurrentProvider<'a>;

    fn config (&self) -> &Config;
    fn make_current (&'a self, surface: SurfaceRef<'a>) -> Result<Self::Current>;
    fn new (builder: &ContextBuilder) -> Result<Self>;
}

/// OpenGL context.
pub struct Context {
    provider: ContextProvider,
}

impl Context {
    pub fn config (&self) -> &Config { self.provider.config() }

    pub fn with<'a, S, F, R> (&'a self, surface: S, mut f: F) -> Result<R>
        where S: Into<SurfaceRef<'a>>, F: FnMut(&Current) -> R
    {
        let _lock = ThreadLock::acquire()?;
        let surface = surface.into();
        let current = Current {
            provider: self.provider.make_current(surface)?,
            context: self,
            surface: surface,
        };

        let ret = f(&current);

        if let Err(err) = current.provider.release() {
            error!("failed to release OpenGL context: {}", err);
        }

        Ok(ret)
    }
}

impl GetProvider for Context {
    type Provider = ContextProvider;
    fn provider (&self) -> &ContextProvider { &self.provider }
}

/// `Current` provider trait.
pub trait ICurrentProvider<'a> : Sized {
    fn get_proc_address (&self, name: &str) -> Result<*mut c_void>;
    fn release (self) -> Result<()>;
    fn set_swap_interval (&self, interval: i8) -> Result<()>;
    fn swap_buffers (&self) -> Result<()>;
}

/// Current OpenGL selection.
pub struct Current<'a> {
    provider: CurrentProvider<'a>,
    context: &'a Context,
    surface: SurfaceRef<'a>,
}

impl<'a> Current<'a> {
    pub fn context (&self) -> &'a Context { self.context }

    pub fn get_proc_address (&self, name: &str) -> Result<*mut c_void> {
        self.provider.get_proc_address(name)
    }

    pub fn set_swap_interval (&self, interval: i8) -> Result<()> {
        self.provider.set_swap_interval(interval)
    }

    pub fn surface (&self) -> SurfaceRef<'a> { self.surface }

    pub fn swap_buffers (&self) -> Result<()> {
        self.provider.swap_buffers()
    }

    pub fn version (&self) -> Result<Version> {
        type PFNGLGETSTRINGPROC = unsafe extern "C" fn (c_uint) -> *const c_char;
        const GL_VERSION: c_uint = 0x1f02;

        let get_string: PFNGLGETSTRINGPROC;
        let version_str;

        unsafe {
            get_string = match self.get_proc_address("glGetString") {
                Ok(f) => mem::transmute(f),
                Err(_) => { return Err(err!(NoMatch).with_str("glGetString)")); },
            };

            let version_ptr = get_string(GL_VERSION);

            if version_ptr.is_null() {
                return Err(err!(NoMatch).with_str("GL_VERSION"));
            }

            version_str = CStr::from_ptr(version_ptr);
        }

        parse_version(version_str.to_bytes())
    }
}

impl<'a> GetProvider for Current<'a> {
    type Provider = CurrentProvider<'a>;
    fn provider (&self) -> &CurrentProvider<'a> { &self.provider }
}

/// Manages the `THREAD_LOCKED` flag with RAII.
pub struct ThreadLock {
    _uninstantiable: (),
}

impl ThreadLock {
    pub fn acquire () -> Result<ThreadLock> {
        THREAD_LOCKED.with(|locked| {
            let mut locked_ref = locked.borrow_mut();

            if *locked_ref {
                return Err(err!(InvalidOperation).with_str("nested OpenGL blocks"));
            }

            *locked_ref = true;
            Ok(ThreadLock { _uninstantiable: () })
        })
    }
}

impl Drop for ThreadLock {
    fn drop (&mut self) {
        THREAD_LOCKED.with(|l| { *l.borrow_mut() = false; });
    }
}

/// Returns true if a byte is an ASCII digit.
fn is_digit (byte: u8) -> bool {
    match byte {
        b'0'...b'9' => true,
        _ => false
    }
}

/// Parses an OpenGL version string.
fn parse_version (s: &[u8]) -> Result<Version> {
    let len = s.len();
    let mut pos = 0;
    let mut major_str = &s[..0];

    while pos < len && is_digit(s[pos]) {
        pos += 1;
        major_str = &s[..pos];
    }

    if pos == len || s[pos] != b'.' {
        return Err(err!(FormatError));
    }

    pos += 1;
    let start = pos;
    let mut minor_str = &s[start..pos];

    while pos < len && is_digit(s[pos]) {
        pos += 1;
        minor_str = &s[start..pos];
    }

    if major_str.is_empty() || minor_str.is_empty() {
        return Err(err!(FormatError));
    }

    match (u8::from_str(str::from_utf8(major_str).unwrap()),
           u8::from_str(str::from_utf8(minor_str).unwrap()))
    {
        (Ok(major), Ok(minor)) => Ok(Version(major, minor)),
        _ => Err(err!(FormatError)),
    }
}
