// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::os::raw::*;
use std::rc::Rc;
use std::slice;
use std::sync::Arc;
use std::vec;

use dwindow::{Display, Visual};
use dwindow::sys::freedesktop::DisplayExt as _DisplayExt0;
use dwindow::sys::freedesktop::VisualExt as _VisualExt0;
use dwindow::sys::x11::Connection;
use try_from::TryFrom;
use x11::{glx, xlib};

use config::{ConfigInfo, IConfigProvider, IConfigsProvider, PixelFormat, RgbFormat};
use error::Result;
use util::GetRef;

/// GLX `Config` provider.
#[derive(Clone)]
pub struct ConfigProvider {
    ctn: Rc<Connection>,
    glx: Arc<glx::Glx>,
    display_ptr: *mut xlib::Display,
    visual: Visual,
    fbconfig_ptr: glx::GLXFBConfig,
}

impl ConfigProvider {
    pub fn attr (&self, attr: c_int) -> Result<c_int> {
        let mut value = 0;

        unsafe {
            if (self.glx.glXGetFBConfigAttrib)(self.display_ptr,
                                               self.fbconfig_ptr,
                                               attr,
                                               &mut value)
                                               != 0
            {
                return Err(err!(RequestFailed).with_str("glXGetFBConfigAttrib"));
            }
        }

        Ok(value)
    }

    pub fn fbconfig_ptr (&self) -> glx::GLXFBConfig { self.fbconfig_ptr }

    pub unsafe fn from_fbconfig_ptr<T> (src: &T,
                                        fbconfig_ptr: glx::GLXFBConfig)
                                        -> Result<ConfigProvider>
        where T: GetRef<Rc<Connection>> + GetRef<Arc<glx::Glx>>
    {
        assert!(!fbconfig_ptr.is_null());

        let ctn: Rc<Connection> = src.get_clone();
        let xlib = ctn.xlib().clone();
        let glx: Arc<glx::Glx> = src.get_clone();
        let display_ptr = ctn.display_ptr();
        let visual_info_ptr = (glx.glXGetVisualFromFBConfig)(display_ptr, fbconfig_ptr);

        if visual_info_ptr.is_null() {
            return Err(err!(RequestFailed).with_str("glXGetVisualFromFBConfig"));
        }

        let visual_info = *visual_info_ptr;
        (xlib.XFree)(visual_info_ptr as *mut _);
        let visual = Visual::from_x_visual_info(ctn.clone(), visual_info);

        Ok(ConfigProvider {
            ctn: ctn,
            glx: glx,
            display_ptr: display_ptr,
            visual: visual,
            fbconfig_ptr: fbconfig_ptr,
        })
    }
}

impl Eq for ConfigProvider {}

impl GetRef<Rc<Connection>> for ConfigProvider {
    fn get_ref (&self) -> &Rc<Connection> { &self.ctn }
}

impl GetRef<Arc<glx::Glx>> for ConfigProvider {
    fn get_ref (&self) -> &Arc<glx::Glx> { &self.glx }
}

impl IConfigProvider for ConfigProvider {
    fn info (&self) -> Result<ConfigInfo> {
        let drawable_type = self.attr(glx::GLX_DRAWABLE_TYPE)?;
        let render_type = self.attr(glx::GLX_RENDER_TYPE)?;

        Ok(ConfigInfo {
            support_gl: true,
            support_gles: false,
            draw_to_window: drawable_type & glx::GLX_WINDOW_BIT != 0,
            draw_to_pixmap: drawable_type & glx::GLX_PIXMAP_BIT != 0,
            pixel_format: if render_type & glx::GLX_RGBA_BIT != 0 {
                PixelFormat::Rgb(RgbFormat {
                    red_size: u8::try_from(self.attr(glx::GLX_RED_SIZE)?)?,
                    green_size: u8::try_from(self.attr(glx::GLX_GREEN_SIZE)?)?,
                    blue_size: u8::try_from(self.attr(glx::GLX_BLUE_SIZE)?)?,
                    alpha_size: u8::try_from(self.attr(glx::GLX_ALPHA_SIZE)?)?,
                    accum_red_size: u8::try_from(self.attr(glx::GLX_ACCUM_RED_SIZE)?)?,
                    accum_green_size: u8::try_from(self.attr(glx::GLX_ACCUM_GREEN_SIZE)?)?,
                    accum_blue_size: u8::try_from(self.attr(glx::GLX_ACCUM_BLUE_SIZE)?)?,
                    accum_alpha_size: u8::try_from(self.attr(glx::GLX_ACCUM_ALPHA_SIZE)?)?,
                })
            } else if render_type & glx::GLX_COLOR_INDEX_BIT != 0 {
                PixelFormat::Index(u8::try_from(self.attr(glx::GLX_BUFFER_SIZE)?)?)
            } else {
                return Err(err!(Unsupported).with_str("GLX_RENDER_TYPE"));
            },
            depth_size: u8::try_from(self.attr(glx::GLX_DEPTH_SIZE)?)?,
            stencil_size: u8::try_from(self.attr(glx::GLX_STENCIL_SIZE)?)?,
            double_buffer: self.attr(glx::GLX_DOUBLEBUFFER)? != 0,
            stereoscopic: self.attr(glx::GLX_STEREO)? != 0,
        })
    }

    fn visual (&self) -> &Visual { &self.visual }
}

impl PartialEq for ConfigProvider {
    fn eq (&self, rhs: &ConfigProvider) -> bool {
        self.fbconfig_ptr == rhs.fbconfig_ptr
    }
}

/// GLX `Configs` provider.
pub struct ConfigsProvider {
    ctn: Rc<Connection>,
    glx: Arc<glx::Glx>,
    inner: vec::IntoIter<glx::GLXFBConfig>,
}

impl ConfigsProvider {
    pub fn gl_configs (display: &Display) -> Result<ConfigsProvider> {
        let display_raw = display.as_raw().try_x11()?;
        let ctn = display_raw.connection.clone();
        let xlib = ctn.xlib().clone();
        let glx = ::provider::glx::open_glx()?;
        let display_ptr = ctn.display_ptr();
        let screen_num = display_raw.screen;
        let fbconfigs;

        unsafe {
            let mut n_items = 0;
            let atts = [0];
            let fbconfig_list = (glx.glXChooseFBConfig)(display_ptr,
                                                        screen_num,
                                                        atts.as_ptr(),
                                                        &mut n_items);

            if fbconfig_list.is_null() {
                return Err(err!(RequestFailed).with_str("glXChooseFBConfig"));
            } else if n_items < 1 {
                (xlib.XFree)(fbconfig_list as *mut _);
                return Err(err!(RequestFailed).with_str("glXChooseFBConfig"));
            }

            let n_items = usize::try_from(n_items).unwrap();
            let fbconfig_slice = slice::from_raw_parts(fbconfig_list as *const glx::GLXFBConfig,
                                                       n_items);
            fbconfigs = Vec::from(fbconfig_slice);
            (xlib.XFree)(fbconfig_list as *mut _);
        }

        Ok(ConfigsProvider {
            ctn: ctn,
            glx: glx,
            inner: fbconfigs.into_iter(),
        })
    }
}

impl GetRef<Rc<Connection>> for ConfigsProvider {
    fn get_ref (&self) -> &Rc<Connection> { &self.ctn }
}

impl GetRef<Arc<glx::Glx>> for ConfigsProvider {
    fn get_ref (&self) -> &Arc<glx::Glx> { &self.glx }
}

impl IConfigsProvider for ConfigsProvider {
    type Config = ConfigProvider;

    fn next (&mut self) -> Result<Option<ConfigProvider>> {
        let fbconfig_ptr = match self.inner.next() {
            None => { return Ok(None); },
            Some(p) => p,
        };

        unsafe {
            Ok(Some(ConfigProvider::from_fbconfig_ptr(self, fbconfig_ptr)?))
        }
    }
}
