// Copyright (c) 2017, Marty Mills <daggerbot@gmail.com>
// This software is available under the terms of the zlib license.
// See COPYING.md for more information.

use std::ffi::CString;
use std::mem;
use std::os::raw::*;
use std::ptr;
use std::rc::Rc;
use std::sync::Arc;

use dwindow::sys::freedesktop::VisualExt as _VisualExt0;
use dwindow::sys::x11::Connection;
use x11::{glx, xlib};

use config::Config;
use context::{Api, ContextBuilder, IContextProvider, ICurrentProvider, Profile};
use display::DisplayExt;
use error::Result;
use provider::GetProvider;
use provider::glx::surface::SurfaceRefExt;
use surface::SurfaceRef;
use util::GetRef;

type PFNGLXCREATECONTEXTATTRIBSARBPROC = extern "C" fn (*mut xlib::Display,
                                                        glx::GLXFBConfig,
                                                        glx::GLXContext,
                                                        xlib::Bool,
                                                        *const c_int)
                                                        -> glx::GLXContext;

type PFNGLXSWAPINTERVALEXTPROC = extern "C" fn (*mut xlib::Display,
                                                glx::GLXDrawable,
                                                c_int);

/// GLX `Context` provider.
pub struct ContextProvider {
    ctn: Rc<Connection>,
    glx: Arc<glx::Glx>,
    display_ptr: *mut xlib::Display,
    screen_num: c_int,
    context_ptr: glx::GLXContext,
    config: Config,
}

impl ContextProvider {
    pub fn new_arb (builder: &ContextBuilder,
                    bs_context: &ContextProvider)
                    -> Result<ContextProvider>
    {
        let ctn: Rc<Connection> = bs_context.ctn.clone();
        let glx: Arc<glx::Glx> = bs_context.glx.clone();
        let display_ptr = bs_context.display_ptr;
        let screen_num = bs_context.screen_num;
        let bs_window = builder.display().new_window()
                                         .with_callback(|_| ()) // Discard events
                                         .with_visual(&bs_context.config)
                                         .build()?;
        let bs_current = bs_context.make_current(SurfaceRef::Window(&bs_window))?;
        let mut attribs = Vec::new();

        if let Some(profile) = builder.profile() {
            attribs.push(glx::arb::GLX_CONTEXT_PROFILE_MASK_ARB);

            attribs.push(match profile {
                Profile::Core => glx::arb::GLX_CONTEXT_CORE_PROFILE_BIT_ARB,
                Profile::Compatibility => glx::arb::GLX_CONTEXT_COMPATIBILITY_PROFILE_BIT_ARB,
            });
        }

        if let Some(version) = builder.version() {
            attribs.push(glx::arb::GLX_CONTEXT_MAJOR_VERSION_ARB);
            attribs.push(version.0 as c_int);
            attribs.push(glx::arb::GLX_CONTEXT_MINOR_VERSION_ARB);
            attribs.push(version.1 as c_int);
        }

        attribs.push(0);

        let share_ptr = match builder.shared_context() {
            None => ptr::null_mut(),
            Some(share) => share.provider().context_ptr,
        };

        let create: PFNGLXCREATECONTEXTATTRIBSARBPROC;
        let context_ptr;

        unsafe {
            let fptr = match bs_current.get_proc_address("glXCreateContextAttribsARB") {
                Ok(p) => p,
                Err(err) => {
                    return Err(err!(NoMatch).with_str("glXCreateContextAttribsARB")
                                            .with_cause(err));
                },
            };

            create = mem::transmute(fptr);
            context_ptr = create(display_ptr,
                                 bs_context.config.provider().fbconfig_ptr(),
                                 share_ptr,
                                 xlib::True,
                                 attribs.as_ptr());
        }

        if context_ptr.is_null() {
            return Err(err!(RequestFailed).with_str("glXCreateContextAttribsARB"));
        }

        Ok(ContextProvider {
            ctn: ctn,
            glx: glx,
            display_ptr: display_ptr,
            screen_num: screen_num,
            context_ptr: context_ptr,
            config: bs_context.config.clone(),
        })
    }

    pub fn new_legacy (config: &Config,
                       share: Option<&ContextProvider>)
                       -> Result<ContextProvider>
    {
        let ctn: Rc<Connection> = config.provider().get_clone();
        let glx: Arc<glx::Glx> = config.provider().get_clone();
        let display_ptr = ctn.display_ptr();
        let mut visual_info = config.visual().as_raw().try_x11()?.visual_info;
        let share_ptr = match share {
            None => ptr::null_mut(),
            Some(share) => {
                if share.ctn != ctn || share.screen_num != visual_info.screen {
                    return Err(err!(IncompatibleResource).with_str("share context"));
                }
                share.context_ptr
            },
        };
        let context_ptr;

        unsafe {
            context_ptr = (glx.glXCreateContext)(display_ptr,
                                                 &mut visual_info,
                                                 share_ptr,
                                                 xlib::True);
        }

        if context_ptr.is_null() {
            return Err(err!(RequestFailed).with_str("glXCreateContext"));
        }

        Ok(ContextProvider {
            ctn: ctn,
            glx: glx,
            display_ptr: display_ptr,
            screen_num: visual_info.screen,
            context_ptr: context_ptr,
            config: config.clone(),
        })
    }
}

impl Drop for ContextProvider {
    fn drop (&mut self) {
        unsafe {
            (self.glx.glXDestroyContext)(self.display_ptr, self.context_ptr);
        }
    }
}

impl Eq for ContextProvider {}

impl GetRef<Rc<Connection>> for ContextProvider {
    fn get_ref (&self) -> &Rc<Connection> { &self.ctn }
}

impl GetRef<Arc<glx::Glx>> for ContextProvider {
    fn get_ref (&self) -> &Arc<glx::Glx> { &self.glx }
}

impl<'a> IContextProvider<'a> for ContextProvider {
    type Current = CurrentProvider<'a>;

    fn config (&self) -> &Config { &self.config }

    fn make_current (&'a self, surface: SurfaceRef<'a>) -> Result<CurrentProvider<'a>> {
        let drawable = surface.try_xid()?;

        unsafe {
            if (self.glx.glXMakeCurrent)(self.display_ptr, drawable, self.context_ptr) == 0 {
                return Err(err!(RequestFailed).with_str("glXMakeCurrent"));
            }
        }

        Ok(CurrentProvider {
            glx: self.glx.clone(),
            display_ptr: self.display_ptr,
            _context: self,
            surface: surface,
        })
    }

    fn new (builder: &ContextBuilder) -> Result<ContextProvider> {
        match builder.api() {
            Api::Opengl => (),
            Api::Opengles => { return Err(err!(Unimplemented).with_str("OpenGL ES context")); },
        }

        let _thread_lock = ::context::ThreadLock::acquire()?;

        let config = match builder.config() {
            None => match builder.shared_context() {
                None => builder.display().default_gl_config()?,
                Some(share) => share.provider().config.clone(),
            },
            Some(config) => config.clone(),
        };
        let visual_raw = config.visual().as_raw().try_x11()?;

        let share = match builder.shared_context() {
            None => None,
            Some(share) => {
                if share.provider().ctn != *visual_raw.connection
                   || share.provider().screen_num != visual_raw.visual_info.screen
                {
                    return Err(err!(IncompatibleResource).with_str("share context"));
                }
                Some(share.provider())
            },
        };

        let bs_context = ContextProvider::new_legacy(&config, share)?;

        match ContextProvider::new_arb(builder, &bs_context) {
            Ok(c) => Ok(c),
            Err(err) => {
                error!("can't create ARB context: {}", err);
                Ok(bs_context)
            },
        }
    }
}

impl PartialEq for ContextProvider {
    fn eq (&self, rhs: &ContextProvider) -> bool {
        self as *const _ == rhs as *const _
    }
}

/// GLX `Current` provider.
pub struct CurrentProvider<'a> {
    glx: Arc<glx::Glx>,
    display_ptr: *mut xlib::Display,
    _context: &'a ContextProvider,
    surface: SurfaceRef<'a>,
}

impl<'a> ICurrentProvider<'a> for CurrentProvider<'a> {
    fn get_proc_address (&self, name: &str) -> Result<*mut c_void> {
        let c_name = CString::new(name)?;

        unsafe {
            let ptr = (self.glx.glXGetProcAddress)(c_name.as_bytes().as_ptr());

            if ptr.is_none() {
                return Err(err!(NoMatch).with_str("glXGetProcAddress"));
            }

            Ok(mem::transmute(ptr))
        }
    }

    fn release (self) -> Result<()> {
        unsafe {
            if (self.glx.glXMakeCurrent)(self.display_ptr, 0, ptr::null_mut()) == 0 {
                return Err(err!(RequestFailed).with_str("glXMakeCurrent"));
            }
        }

        Ok(())
    }

    fn set_swap_interval (&self, interval: i8) -> Result<()> {
        match self.surface {
            // Should return an error when used on pixmaps
            SurfaceRef::Window(_) => (),
        }

        let drawable = self.surface.try_xid()?;
        let set: PFNGLXSWAPINTERVALEXTPROC;

        unsafe {
            let ptr = match self.get_proc_address("glXSwapIntervalEXT") {
                Ok(p) => p,
                Err(_) => { return Err(err!(NoMatch).with_str("glXSwapIntervalEXT")); },
            };

            set = mem::transmute(ptr);
            set(self.display_ptr, drawable, interval.abs() as c_int);
        }

        Ok(())
    }

    fn swap_buffers (&self) -> Result<()> {
        let drawable = self.surface.try_xid()?;

        unsafe {
            (self.glx.glXSwapBuffers)(self.display_ptr, drawable);
        }

        Ok(())
    }
}
